/*********************************************************************
 Copyright (C) 2003 Imperial College London

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/

package vishnu.server;

import vishnu.server.Search.*;
import vishnu.datablock.*;
import vishnu.util.XMLUtil;

import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;
import java.util.Vector;
import java.util.Hashtable;

import org.w3c.dom.Node; 
import org.w3c.dom.NodeList; 
import org.w3c.dom.Document; 
import org.w3c.dom.Element; 
import org.xml.sax.InputSource;
import javax.xml.parsers.*;

public class VisServlet extends HttpServlet
{
    final String DOCUMENT_FETCH_ACTION = "fetch";
    final String DOCUMENT_SEARCH_ACTION = "search";
    final String KEYWORDS_COMPUTE_ACTION = "keywords";
    final String INITIALIZATION_ACTION = "initialise";
    
    protected String COLLS_HOME=null; 
    
    Hashtable collH;

    public void init(ServletConfig config) throws ServletException {
	// always call super.init;
	super.init(config);

	COLLS_HOME = config.getInitParameter("collectionsHome");
	if (COLLS_HOME == null || COLLS_HOME.equals("")) {
	    System.err.println("initialization parameter, collectionsHome, not set");
	    System.exit(1);
	}
	if (!COLLS_HOME.endsWith(File.separator)) {
	    COLLS_HOME += File.separator;
	}
    }

    private Object receiveObject(HttpServletRequest req) throws Exception
    {
        ObjectInputStream in = new ObjectInputStream( req.getInputStream() );
        Object obj = in.readObject();
        in.close();
        return obj;
    }
    
    
    private void sendObject(HttpServletResponse resp, Object obj) throws Exception
    {
        ObjectOutputStream out = new ObjectOutputStream( resp.getOutputStream() );
        out.writeObject(obj);
        out.close();
    }
    
    public void doGet(HttpServletRequest req, HttpServletResponse resp) 
	throws IOException {

	resp.setContentType("text/html;charset=UTF-8");
        PrintWriter out = resp.getWriter();
	String output = "<html><head><title>Visualiser servlet</title></head>\n<body>The servlet is running, and accepting messages via POST</body></html>";
	out.println(output);

    }
    public void doPost(HttpServletRequest req, HttpServletResponse resp)
    {
	Object obj;
	
        try{
	    obj = receiveObject(req);
	    Vector q = (Vector)obj;
	    Vector response = new Vector();
	    
	    String action = (String)q.elementAt(0);
	    System.out.println("Action.. " + action);
	    
 	    if( action.equals(INITIALIZATION_ACTION) ){
		Vector colls = activateCollections();
		sendObject(resp,colls);
	    }
	    else if( action.equals(DOCUMENT_SEARCH_ACTION) ){
		Vector docNums = new Vector();
		Vector descriptions = new Vector();
		String engine;
		
		String query = (String)q.elementAt(1);
		String collection = (String)q.elementAt(2);
		String view = null;
		if (q.size() == 4) {
		    view = (String)q.elementAt(3);
		}
		if (view != null && view.equals("")) {
		    view = null;
		}
		System.out.println("Requesting..." + collection);
		
		try{
		    engine = (String)collH.get(collection);
		}catch(Exception ex){engine = "MG";}
		
		
		System.out.println("Retrieving with " + engine);
		
		String collection_home = COLLS_HOME+collection+File.separator;
		SearchInterface search_engine = null;
		if( engine.equals("MG") ){
		    search_engine = new MGSearcher(collection_home, collection, view);
		} else if( engine.equals("LUCENE") ){
		    search_engine = new LUCSearcher(collection_home, view);
		} 
		
		search_engine.search(query);
		docNums = search_engine.getDocIdentifiers();
		descriptions = search_engine.getDocDescriptions();
		System.out.println(docNums.size() + " documents in VisServlet");
		DataBlock data = generateDataBlock(collection_home, view, docNums, descriptions);
		if (data != null) {
		    response.addElement(data);
		}
		sendObject(resp,response);
		search_engine = null;
		data = null;
		docNums = null;
		descriptions = null;
		
	    }
	    else if( action.equals(DOCUMENT_FETCH_ACTION) ){
		String docNum = (String)q.elementAt(1);
		String collection = (String)q.elementAt(2);
		String view = null;
		if (q.size() == 4) {
		    view = (String)q.elementAt(3);
		}
		if (view != null && view.equals("")) {
		    view = null;
		}
		String engine;
		String collection_home = COLLS_HOME+collection+File.separator;
		
		try{
		    engine = (String)collH.get(collection);
		}catch(Exception ex){engine = "MG";}
		
                System.out.println("Retrieving with " + engine);
		
		SearchInterface search_engine = null;
		if( engine.equals("MG") ){
		    search_engine = new MGSearcher(collection_home, collection, view);
		} else if( engine.equals("LUCENE") ){
		    search_engine = new LUCSearcher(collection_home, view);
		} 
		System.err.println("Searching for doc "+docNum);
		Vector doc = search_engine.getDocContent(Integer.parseInt(docNum));
		response.addElement(doc);			
		sendObject(resp,response);
		search_engine = null;
	    }
	} catch(Exception e){}
    }
    

    public Vector connect(Vector q)
    {
	return null;
    }


    private Vector activateCollections()
    {
	System.err.println("activating collections");
	Vector colls = new Vector();
	collH = new Hashtable();
	Document doc = null;
	
	File dir = new File(COLLS_HOME);
	if( !dir.exists() ){
	    System.err.println("collections directory, "+COLLS_HOME+", doesn't exist, can't activate collections");
	    return null;
	}
	
	File[] list = dir.listFiles();
	
	for( int i = 0; i < list.length; i++ ){
	    File f = list[i];
	    if( f.isDirectory()){
		File config_file = new File(f, "config.xml");
		if (!config_file.exists()) {
		    System.err.println("Collection "+f.getName()+" has no config file, ignoring");
		    // ignore this collection
		    continue;
		} else {
		    System.err.println("Trying to init collection "+f.getName());
		}
		try {
		    DocumentBuilderFactory doc_build_fac = DocumentBuilderFactory.newInstance();
		    DocumentBuilder doc_build = doc_build_fac.newDocumentBuilder();
		    doc = doc_build.parse(config_file);
		} catch (Exception e) {
		    System.err.println("Error parsing config file "+config_file.getPath()+", ignoring this collection");
		    continue;
		}
		Element root = doc.getDocumentElement();
		String name = root.getAttribute("name");
		String desc = root.getAttribute("description");
		String engine = root.getAttribute("engine");
		int size = Integer.parseInt(root.getAttribute("size"));
		String coll_id = f.getName();
		Element view_elem = (Element)XMLUtil.getChildByTagName(root, "viewList");
		String [] views = null;
		if (view_elem != null) {
		    NodeList view_list = view_elem.getElementsByTagName("view");
		    int num_views = view_list.getLength();
		    if (num_views > 0) {
			views = new String[num_views];
			for (int j=0; j<num_views; j++) {
			    String v = ((Element)view_list.item(j)).getAttribute("name");
			    views[j]=v;
			}
		    }
		}
		Collection c = new Collection(coll_id,name,desc,size,engine,views);
		System.out.println("Hashing.." + coll_id);
		colls.addElement(c);
		collH.put(coll_id,engine);
	    }			
	}
	doc = null;
	return colls;
    }
    
    private DataBlock generateDataBlock(String collection_home, String view, Vector docNums, Vector descriptions) {

	if (docNums.size()==0) {
	    return null;
	}
	DataBlock data = new DataBlock();

	CKServer ck_server = new CKServer(collection_home, view);
	
	/**** set data fields one by one and pass one what ever gets assembled ****/
		
	try{
	    
	    ck_server.setDescriptions(descriptions);
	    
	    /**** get candidate keywords ****/
	    
	    String[] keywords = ck_server.computeKeywords(docNums);
	    
	    data.words = keywords;
	    
	    
	    /**** get sparse document * keyword matrix ****/
	    
	    SparseMatrix matrix = ck_server.getSparseMatrix(docNums);
	    
	    data.matrix = matrix;
	    
	    
	    /**** get document indices, this is a subset of the original ****/
	    /**** those without keywords are excluded                    ****/
	    
	    int[] docs = ck_server.getHitDocuments();
	    
	    data.docs = docs;
	    
	    
	    String[] desc = ck_server.getHitDescriptions();
	    
	    data.descriptions = desc;
	    
	    
	    /**** get 10 or so clusters ****/
	    
	    Vector[] clusters = ck_server.getClusters();
	    
	    data.clusters = clusters;
	    
	    
	    double[][] centroids = ck_server.getCentroids();
	    
	    
	    /**** send their centroids through sammon mapping ****/
	    
	    Point2D[] sammon = ck_server.getSammonMap(centroids);
	    
	    System.out.println("Num descriptions: " + descriptions.size());
	    System.out.println("Num docs: " + docs.length);
	    System.out.println("Num desc: " + desc.length);
	    
	    data.sammon = sammon;
	    
	} catch (Exception e) {
	    System.err.println("VisServlet: computing clustering Error: "+e);
	    e.printStackTrace();
	}
	ck_server = null;
	return data;
    }
}





